--- bcm2835_sdhci.c-20130110	2013-01-10 05:26:42.000000000 +0900
+++ bcm2835_sdhci.c	2013-01-10 23:49:13.000000000 +0900
@@ -302,6 +302,13 @@
 		bcm2835_asm_sdhci_write_4(slot->bus, slot, SDHCI_INT_STATUS,
 					  SDHCI_INT_DATA_END);
 
+		/* drain write buffer if READ case */
+		if (slot->curcmd->data->flags & MMC_DATA_READ) {
+			isb(); dsb(); /* memory barrier */
+			bcm2835_dma_wb_range(slot->curcmd->data->data,
+					     slot->curcmd->data->len);
+		}
+
 		/* enable INT */
 		slot->intmask |= SDHCI_INT_DATA_AVAIL | SDHCI_INT_SPACE_AVAIL
 				| SDHCI_INT_DATA_END;
@@ -586,6 +593,12 @@
 			    & SDHCI_DATA_AVAILABLE) {
 				bcm_sdhci_read_pio(slot);
 			}
+			/* drain write buffer if READ && DMA+PIO case */
+			if (slot->curcmd->data->len >= 512) {
+				isb(); dsb(); /* memory barrier */
+				bcm2835_dma_wb_range(slot->curcmd->data->data,
+						     slot->curcmd->data->len);
+			}
 		} else {
 			if (bcm2835_asm_sdhci_read_4(bus, slot, SDHCI_PRESENT_STATE)
 			    & SDHCI_SPACE_AVAILABLE) {
@@ -602,9 +615,14 @@
 
 	/* Now use DMA transfer */
 #ifdef DMA_DEBUG
-	printf("DMA transfer (left=%d,len=%d)\n", left, slot->curcmd->data->len);
-	reg = bcm2835_asm_sdhci_read_4(bus, slot, SDHCI_INT_STATUS);
-	printf("reg=%x\n", reg);
+	{
+		volatile uint32_t reg;
+		printf("DMA transfer %s (left=%d,len=%d)\n",
+		       (slot->curcmd->data->flags & MMC_DATA_READ) ? "read" : "write",
+		       left, slot->curcmd->data->len);
+		reg = bcm2835_asm_sdhci_read_4(bus, slot, SDHCI_INT_STATUS);
+		printf("reg=%x\n", reg);
+	}
 #endif
 
 	/* Disable INT */
